@@ -51,6 +51,7 @@
 #include <getopt.h>
 const struct option longopts[] = {
     {"force", 0, 0, 'f'},
+    {"really-force", 0, 0, 'r'},
     {"system-only", 0, 0, 's'},
     {"version", 0, 0, 'V'},
     {"verbose", 0, 0, 'v'},
@@ -68,10 +69,10 @@ static void
 usage (char *program)
 {
 #if HAVE_GETOPT_LONG
-    fprintf (stderr, "usage: %s [-fsvV?] [--force] [--system-only] [--verbose] [--version] [--help] [dirs]\n",
+    fprintf (stderr, "usage: %s [-frsvV?] [--force|--really-force] [--system-only] [--verbose] [--version] [--help] [dirs]\n",
 	     program);
 #else
-    fprintf (stderr, "usage: %s [-fsvV?] [dirs]\n",
+    fprintf (stderr, "usage: %s [-frsvV?] [dirs]\n",
 	     program);
 #endif
     fprintf (stderr, "Build font information caches in [dirs]\n"
@@ -79,12 +80,14 @@ usage (char *program)
     fprintf (stderr, "\n");
 #if HAVE_GETOPT_LONG
     fprintf (stderr, "  -f, --force          scan directories with apparently valid caches\n");
+    fprintf (stderr, "  -r, --really-force   erase all existing caches, then rescan\n");
     fprintf (stderr, "  -s, --system-only    scan system-wide directories only\n");
     fprintf (stderr, "  -v, --verbose        display status information while busy\n");
     fprintf (stderr, "  -V, --version        display font config version and exit\n");
     fprintf (stderr, "  -?, --help           display this help and exit\n");
 #else
     fprintf (stderr, "  -f         (force)   scan directories with apparently valid caches\n");
+    fprintf (stderr, "  -r,   (really force) erase all existing caches, then rescan\n");
     fprintf (stderr, "  -s         (system)  scan system-wide directories only\n");
     fprintf (stderr, "  -v         (verbose) display status information while busy\n");
     fprintf (stderr, "  -V         (version) display font config version and exit\n");
@@ -93,6 +96,8 @@ usage (char *program)
     exit (1);
 }
 
+static FcStrSet *processed_dirs;
+
 static int
 nsubdirs (FcStrSet *set)
 {
@@ -109,10 +114,11 @@ nsubdirs (FcStrSet *set)
 }
 
 static int
-scanDirs (FcStrList *list, FcConfig *config, char *program, FcBool force, FcBool verbose)
+scanDirs (FcStrList *list, FcConfig *config, char *program, FcBool force, FcBool really_force, FcBool verbose)
 {
     int		ret = 0;
-    FcChar8	*dir;
+    const FcChar8 *dir;
+    const FcChar8 *dir_orig;
     FcFontSet	*set;
     FcStrSet	*subdirs;
     FcStrList	*sublist;
@@ -122,13 +128,30 @@ scanDirs (FcStrList *list, FcConfig *config, char *program, FcBool force, FcBool
      * Now scan all of the directories into separate databases
      * and write out the results
      */
-    while ((dir = FcStrListNext (list)))
+    while ((dir_orig = FcStrListNext (list)))
     {
+	dir = FcConfigNormalizeFontDir (config, dir_orig);
+	
 	if (verbose)
 	{
-	    printf ("%s: \"%s\": ", program, dir);
+	    printf ("%s: \"%s\": ", program, dir ? dir : dir_orig);
 	    fflush (stdout);
 	}
+	
+	if (!dir)
+	{
+	    if (verbose)
+		printf ("skipping, no such directory\n");
+	    continue;
+	}
+	
+	if (FcStrSetMember (processed_dirs, dir))
+	{
+	    if (verbose)
+		printf ("skipping, looped directory detected\n");
+	    continue;
+	}
+
 	set = FcFontSetCreate ();
 	if (!set)
 	{
@@ -141,6 +164,7 @@ scanDirs (FcStrList *list, FcConfig *config, char *program, FcBool force, FcBool
 	{
 	    fprintf (stderr, "Can't create directory set\n");
 	    ret++;
+	    FcFontSetDestroy (set);
 	    continue;
 	}
 	
@@ -151,38 +175,54 @@ scanDirs (FcStrList *list, FcConfig *config, char *program, FcBool force, FcBool
 	    case ENOTDIR:
 		if (verbose)
 		    printf ("skipping, no such directory\n");
-		break;
+		FcFontSetDestroy (set);
+		FcStrSetDestroy (subdirs);
+		continue;
 	    case EACCES:
 	    case EROFS:
-		if (verbose)
-		    printf ("skipping, no write access\n");
+		/* That's ok, caches go to /var anyway. */
+		/* Ideally we'd do an access on the hashed_name. */
+		/* But we hid that behind an abstraction barrier. */
 		break;
 	    default:
 		fprintf (stderr, "\"%s\": ", dir);
 		perror ("");
 		ret++;
+
+		FcFontSetDestroy (set);
+		FcStrSetDestroy (subdirs);
+		continue;
 	    }
-	    continue;
 	}
 	if (stat ((char *) dir, &statb) == -1)
 	{
 	    fprintf (stderr, "\"%s\": ", dir);
 	    perror ("");
+	    FcFontSetDestroy (set);
+	    FcStrSetDestroy (subdirs);
 	    ret++;
 	    continue;
 	}
 	if (!S_ISDIR (statb.st_mode))
 	{
 	    fprintf (stderr, "\"%s\": not a directory, skipping\n", dir);
+	    FcFontSetDestroy (set);
+	    FcStrSetDestroy (subdirs);
 	    continue;
 	}
-	if (!FcDirScan (set, subdirs, 0, FcConfigGetBlanks (config), dir, force))
+
+	if (really_force)
+	    FcDirCacheUnlink (dir, config);
+
+	if (!FcDirScanConfig (set, subdirs, 0, FcConfigGetBlanks (config), dir, force, config))
 	{
 	    fprintf (stderr, "\"%s\": error scanning\n", dir);
+	    FcFontSetDestroy (set);
+	    FcStrSetDestroy (subdirs);
 	    ret++;
 	    continue;
 	}
-	if (!force && FcDirCacheValid (dir))
+	if (!force && FcDirCacheValid (dir) && FcDirCacheHasCurrentArch (dir))
 	{
 	    if (verbose)
 		printf ("skipping, %d fonts, %d dirs\n",
@@ -193,22 +233,32 @@ scanDirs (FcStrList *list, FcConfig *config, char *program, FcBool force, FcBool
 	    if (verbose)
 		printf ("caching, %d fonts, %d dirs\n", 
 			set->nfont, nsubdirs (subdirs));
+
+	    /* This is the only reason we can't combine 
+	     * Valid w/HasCurrentArch... */
+            if (!FcDirCacheValid (dir))
+                if (!FcDirCacheUnlink (dir, config))
+                    ret++;
+
 	    if (!FcDirSave (set, subdirs, dir))
 	    {
-		fprintf (stderr, "Can't save cache in \"%s\"\n", dir);
+                if (!ret)
+                    fprintf (stderr, "Caches are currently saved to \"%s\"\n", PKGCACHEDIR);
+		fprintf (stderr, "Can't save cache for \"%s\"\n", dir);
 		ret++;
 	    }
 	}
 	FcFontSetDestroy (set);
 	sublist = FcStrListCreate (subdirs);
+	FcStrSetDestroy (subdirs);
 	if (!sublist)
 	{
 	    fprintf (stderr, "Can't create subdir list in \"%s\"\n", dir);
 	    ret++;
 	    continue;
 	}
-	ret += scanDirs (sublist, config, program, force, verbose);
-	FcStrSetDestroy (subdirs);
+	FcStrSetAdd (processed_dirs, dir);
+	ret += scanDirs (sublist, config, program, force, really_force, verbose);
     }
     FcStrListDone (list);
     return ret;
@@ -221,6 +271,7 @@ main (int argc, char **argv)
     FcStrList	*list;
     FcBool    	verbose = FcFalse;
     FcBool	force = FcFalse;
+    FcBool	really_force = FcFalse;
     FcBool	systemOnly = FcFalse;
     FcConfig	*config;
     int		i;
@@ -229,12 +280,15 @@ main (int argc, char **argv)
     int		c;
 
 #if HAVE_GETOPT_LONG
-    while ((c = getopt_long (argc, argv, "fsVv?", longopts, NULL)) != -1)
+    while ((c = getopt_long (argc, argv, "frsVv?", longopts, NULL)) != -1)
 #else
-    while ((c = getopt (argc, argv, "fsVv?")) != -1)
+    while ((c = getopt (argc, argv, "frsVv?")) != -1)
 #endif
     {
 	switch (c) {
+	case 'r':
+	    really_force = FcTrue;
+	    /* fall through */
 	case 'f':
 	    force = FcTrue;
 	    break;
@@ -265,6 +319,8 @@ main (int argc, char **argv)
 	fprintf (stderr, "%s: Can't init font config library\n", argv[0]);
 	return 1;
     }
+    FcConfigSetCurrent (config);
+
     if (argv[i])
     {
 	dirs = FcStrSetCreate ();
@@ -288,7 +344,16 @@ main (int argc, char **argv)
     }
     else
 	list = FcConfigGetConfigDirs (config);
-    ret = scanDirs (list, config, argv[0], force, verbose);
+
+    if ((processed_dirs = FcStrSetCreate()) == NULL) {
+	fprintf(stderr, "Cannot malloc\n");
+	return 1;
+    }
+	
+    ret = scanDirs (list, config, argv[0], force, really_force, verbose);
+
+    FcStrSetDestroy (processed_dirs);
+
     /* 
      * Now we need to sleep a second  (or two, to be extra sure), to make
      * sure that timestamps for changes after this run of fc-cache are later
@@ -296,6 +361,7 @@ main (int argc, char **argv)
      * sleep(3) can't be interrupted by a signal here -- this isn't in the
      * library, and there aren't any signals flying around here.
      */
+    FcConfigDestroy (config);
     sleep (2);
     if (verbose)
 	printf ("%s: %s\n", argv[0], ret ? "failed" : "succeeded");
