@@ -22,6 +22,8 @@
  * PERFORMANCE OF THIS SOFTWARE.
  */
 
+#include <dirent.h>
+#include <sys/types.h>
 #include "fcint.h"
 
 #if defined (_WIN32) && (defined (PIC) || defined (DLL_EXPORT))
@@ -139,11 +141,6 @@ bail0:
     return 0;
 }
 
-typedef struct _FcFileTime {
-    time_t  time;
-    FcBool  set;
-} FcFileTime;
-
 static FcFileTime
 FcConfigNewestFile (FcStrSet *files)
 {
@@ -166,6 +163,19 @@ FcConfigNewestFile (FcStrSet *files)
     return newest;
 }
 
+FcFileTime
+FcConfigModifiedTime (FcConfig *config)
+{
+    if (!config)
+    {
+	FcFileTime v = { 0, FcFalse };
+	config = FcConfigGetCurrent ();
+	if (!config)
+	    return v;
+    }
+    return FcConfigNewestFile (config->configFiles);
+}
+
 FcBool
 FcConfigUptoDate (FcConfig *config)
 {
@@ -247,9 +257,10 @@ FcConfigDestroy (FcConfig *config)
 FcBool
 FcConfigBuildFonts (FcConfig *config)
 {
-    FcFontSet	    *fonts;
+    FcFontSet	    *fonts, *cached_fonts;
     FcGlobalCache   *cache;
     FcStrList	    *list;
+    FcStrSet	    *oldDirs;
     FcChar8	    *dir;
 
     fonts = FcFontSetCreate ();
@@ -260,33 +271,71 @@ FcConfigBuildFonts (FcConfig *config)
     if (!cache)
 	goto bail1;
 
+    oldDirs = FcStrSetCreate ();
+    if (!oldDirs)
+        goto bail2;
+
     if (config->cache)
-	FcGlobalCacheLoad (cache, config->cache);
+	FcGlobalCacheLoad (cache, oldDirs, config->cache, config);
 
-    list = FcConfigGetFontDirs (config);
-    if (!list)
-	goto bail1;
-
-    while ((dir = FcStrListNext (list)))
+    cached_fonts = FcCacheRead(config, cache);
+    if (!cached_fonts)
     {
-	if (FcDebug () & FC_DBG_FONTSET)
-	    printf ("scan dir %s\n", dir);
-	FcDirScanConfig (fonts, config->fontDirs, cache, 
-			 config->blanks, dir, FcFalse, config);
+	list = FcConfigGetFontDirs (config);
+	if (!list)
+	    goto bail2;
+	
+	while ((dir = FcStrListNext (list)))
+	{
+	    if (FcDebug () & FC_DBG_FONTSET)
+		printf ("build scan dir %s\n", dir);
+	    FcDirScanConfig (fonts, config->fontDirs, cache, 
+			     config->blanks, dir, FcFalse, config);
+	}
+	
+	FcStrListDone (list);
+    }
+    else
+    {
+	int i;
+
+        for (i = 0; i < oldDirs->num; i++)
+        {
+	    if (FcDebug () & FC_DBG_FONTSET)
+		printf ("scan dir %s\n", oldDirs->strs[i]);
+	    FcDirScanConfig (fonts, config->fontDirs, cache, 
+			     config->blanks, oldDirs->strs[i], 
+                             FcFalse, config);
+	}
+
+	for (i = 0; i < cached_fonts->nfont; i++)
+	{
+	    FcChar8 	*cfn; 
+	    FcPatternGetString (cached_fonts->fonts[i], FC_FILE, 0, &cfn);
+
+	    if (FcConfigAcceptFont (config, cached_fonts->fonts[i]) &&
+                (cfn && FcConfigAcceptFilename (config, cfn)))
+		FcFontSetAdd (fonts, cached_fonts->fonts[i]);
+
+	    cached_fonts->fonts[i] = 0; /* prevent free in FcFontSetDestroy */
+	}
+	cached_fonts->nfont = 0;
+	FcFontSetDestroy (cached_fonts);
     }
-    
-    FcStrListDone (list);
     
     if (FcDebug () & FC_DBG_FONTSET)
 	FcFontSetPrint (fonts);
 
     if (config->cache)
-	FcGlobalCacheSave (cache, config->cache);
+	FcGlobalCacheSave (cache, config->cache, config);
     FcGlobalCacheDestroy (cache);
+    FcStrSetDestroy (oldDirs);
 
     FcConfigSetFonts (config, fonts, FcSetSystem);
     
     return FcTrue;
+bail2:
+    FcStrSetDestroy (oldDirs);
 bail1:
     FcFontSetDestroy (fonts);
 bail0:
@@ -334,13 +383,113 @@ FcConfigGetConfigDirs (FcConfig   *config)
     return FcStrListCreate (config->configDirs);
 }
 
+static FcChar8 *
+FcConfigInodeMatchFontDir (FcConfig *config, const FcChar8 *d)
+{
+    int		n;
+    ino_t	di;
+    dev_t	dd;
+    struct stat s;
+
+    /* first we do string matches rather than file accesses */
+    /* FcStrSetMember doesn't tell us the index so that we can return
+     * the config-owned copy. */
+    for (n = 0; n < config->fontDirs->num; n++)
+    {
+	if (!FcStrCmp (config->fontDirs->strs[n], d))
+	    return config->fontDirs->strs[n];
+    }
+
+    /* If this is a bottleneck, we can cache the fontDir inodes. */
+    if (stat ((char *)d, &s) == -1)
+	return 0;
+    di = s.st_ino; dd = s.st_dev;
+
+    for (n = 0; n < config->fontDirs->num; n++)
+    {
+	if (stat ((char *)config->fontDirs->strs[n], &s) == -1)
+	    continue;
+	if (di == s.st_ino && dd == s.st_dev)
+	    return config->fontDirs->strs[n];
+    }
+    return 0;
+}
+
 FcBool
 FcConfigAddFontDir (FcConfig	    *config,
 		    const FcChar8   *d)
 {
+    /* Avoid adding d if it's an alias of something else, too. */
+    if (FcConfigInodeMatchFontDir(config, d))
+	return FcTrue;
     return FcStrSetAddFilename (config->fontDirs, d);
 }
 
+static FcBool
+FcConfigAddFontDirSubdirs (FcConfig        *config,
+			   const FcChar8   *d)
+{
+    DIR *dir;
+    struct dirent *e;
+    FcChar8 *subdir;
+    FcBool added = FcFalse;
+    
+    if (!(dir = opendir ((char *) d)))
+	return FcFalse;
+    if (!(subdir = (FcChar8 *) malloc (strlen ((char *) d) + FC_MAX_FILE_LEN + 2)))
+    {
+	fprintf (stderr, "out of memory");
+	return FcFalse;
+    }
+    while ((e = readdir (dir)))
+    {
+	if (strcmp (e->d_name, ".") && strcmp (e->d_name, "..") &&
+	    strlen (e->d_name) < FC_MAX_FILE_LEN)
+	{
+	    strcpy ((char *)subdir, (char *)d);
+	    strcat ((char *)subdir, "/");
+	    strcat ((char *)subdir, e->d_name);
+	    if (FcFileIsDir (subdir))
+	    {
+		if (FcConfigInodeMatchFontDir(config, subdir))
+		    continue; /* already added */
+		FcStrSetAddFilename (config->fontDirs, subdir);
+		FcConfigAddFontDirSubdirs (config, subdir);
+		added = FcTrue;
+	    }
+	}
+    }
+    free (subdir);
+    closedir (dir);
+    return added;
+}
+
+const FcChar8 *
+FcConfigNormalizeFontDir (FcConfig  	*config, 
+			  const FcChar8 *d)
+{
+    FcChar8 	*d0;
+    int		n, n0;
+    FcBool 	added = FcFalse;
+
+    d0 = FcConfigInodeMatchFontDir(config, d);
+    if (d0)
+	return d0;
+
+    /* Ok, we didn't find it in fontDirs; let's add subdirs.... */
+    for (n = 0, n0 = config->fontDirs->num; n < n0; n++) 
+    {
+	if (FcConfigAddFontDirSubdirs (config, config->fontDirs->strs[n]))
+	    added = FcTrue;
+    }
+
+    /* ... and try again. */
+    if (added)
+	return FcConfigInodeMatchFontDir(config, d);
+
+    return 0;
+}
+
 FcBool
 FcConfigAddDir (FcConfig	    *config,
 		const FcChar8	    *d)
@@ -562,12 +711,12 @@ FcConfigPromote (FcValue v, FcValue u)
 }
 
 FcBool
-FcConfigCompareValue (const FcValue	left_o,
+FcConfigCompareValue (const FcValue	*left_o,
 		      FcOp		op,
-		      const FcValue	right_o)
+		      const FcValue	*right_o)
 {
-    FcValue	left = left_o;
-    FcValue	right = right_o;
+    FcValue	left = FcValueCanonicalize(left_o);
+    FcValue	right = FcValueCanonicalize(right_o);
     FcBool	ret = FcFalse;
     
     left = FcConfigPromote (left, right);
@@ -629,9 +778,11 @@ FcConfigCompareValue (const FcValue	left_o,
 		ret = FcStrStrIgnoreCase (left.u.s, right.u.s) != 0;
 		break;
 	    case FcOpNotEqual:
-	    case FcOpNotContains:
 		ret = FcStrCmpIgnoreCase (left.u.s, right.u.s) != 0;
 		break;
+	    case FcOpNotContains:
+		ret = FcStrCmpIgnoreCase (left.u.s, right.u.s) == 0;
+		break;
 	    default:
 		break;
 	    }
@@ -753,8 +904,7 @@ FcConfigEvaluate (FcPattern *p, FcExpr *e)
 	break;
     case FcOpString:
 	v.type = FcTypeString;
-	v.u.s = e->u.sval;
-	v = FcValueSave (v);
+	v.u.s = FcStrStaticName(e->u.sval);
 	break;
     case FcOpMatrix:
 	v.type = FcTypeMatrix;
@@ -774,6 +924,7 @@ FcConfigEvaluate (FcPattern *p, FcExpr *e)
 	r = FcPatternGet (p, e->u.field, 0, &v);
 	if (r != FcResultMatch)
 	    v.type = FcTypeVoid;
+	v = FcValueSave (v);
 	break;
     case FcOpConst:
 	if (FcNameConstant (e->u.constant, &v.u.i))
@@ -806,7 +957,7 @@ FcConfigEvaluate (FcPattern *p, FcExpr *e)
 	vl = FcConfigEvaluate (p, e->u.tree.left);
 	vr = FcConfigEvaluate (p, e->u.tree.right);
 	v.type = FcTypeBool;
-	v.u.b = FcConfigCompareValue (vl, e->op, vr);
+	v.u.b = FcConfigCompareValue (&vl, e->op, &vr);
 	FcValueDestroy (vl);
 	FcValueDestroy (vr);
 	break;	
@@ -871,7 +1022,8 @@ FcConfigEvaluate (FcPattern *p, FcExpr *e)
 		switch (e->op) {
 		case FcOpPlus:
 		    v.type = FcTypeString;
-		    v.u.s = FcStrPlus (vl.u.s, vr.u.s);
+		    v.u.s = FcStrStaticName (FcStrPlus (vl.u.s, vr.u.s));
+			 
 		    if (!v.u.s)
 			v.type = FcTypeVoid;
 		    break;
@@ -1019,10 +1171,10 @@ FcConfigMatchValueList (FcPattern	*p,
 	    e = 0;
 	}
 
-	for (v = values; v; v = v->next)
+	for (v = values; v; v = FcValueListPtrU(v->next))
 	{
 	    /* Compare the pattern value to the match expression value */
-	    if (FcConfigCompareValue (v->value, t->op, value))
+	    if (FcConfigCompareValue (&v->value, t->op, &value))
 	    {
 		if (!ret)
 		    ret = v;
@@ -1055,56 +1207,60 @@ FcConfigValues (FcPattern *p, FcExpr *e, FcValueBinding binding)
     if (e->op == FcOpComma)
     {
 	l->value = FcConfigEvaluate (p, e->u.tree.left);
-	l->next  = FcConfigValues (p, e->u.tree.right, binding);
+	l->next  = FcValueListPtrCreateDynamic(FcConfigValues (p, e->u.tree.right, binding));
     }
     else
     {
 	l->value = FcConfigEvaluate (p, e);
-	l->next  = 0;
+	l->next  = FcValueListPtrCreateDynamic(0);
     }
     l->binding = binding;
-    while (l && l->value.type == FcTypeVoid)
+    if (l->value.type == FcTypeVoid)
     {
-	FcValueList	*next = l->next;
-	
+	FcValueList  *next = FcValueListPtrU(l->next);
+
 	FcMemFree (FC_MEM_VALLIST, sizeof (FcValueList));
 	free (l);
 	l = next;
     }
+
     return l;
 }
 
 static FcBool
-FcConfigAdd (FcValueList    **head,
+FcConfigAdd (FcValueListPtr *head,
 	     FcValueList    *position,
 	     FcBool	    append,
 	     FcValueList    *new)
 {
-    FcValueList	    **prev, *last, *v;
+    FcValueListPtr  *prev, last, v;
     FcValueBinding  sameBinding;
     
     if (position)
 	sameBinding = position->binding;
     else
 	sameBinding = FcValueBindingWeak;
-    for (v = new; v; v = v->next)
-	if (v->binding == FcValueBindingSame)
-	    v->binding = sameBinding;
+    for (v = FcValueListPtrCreateDynamic(new); FcValueListPtrU(v); 
+	 v = FcValueListPtrU(v)->next)
+	if (FcValueListPtrU(v)->binding == FcValueBindingSame)
+	    FcValueListPtrU(v)->binding = sameBinding;
     if (append)
     {
 	if (position)
 	    prev = &position->next;
 	else
-	    for (prev = head; *prev; prev = &(*prev)->next)
+	    for (prev = head; FcValueListPtrU(*prev); 
+		 prev = &(FcValueListPtrU(*prev)->next))
 		;
     }
     else
     {
 	if (position)
 	{
-	    for (prev = head; *prev; prev = &(*prev)->next)
+	    for (prev = head; FcValueListPtrU(*prev); 
+		 prev = &(FcValueListPtrU(*prev)->next))
 	    {
-		if (*prev == position)
+		if (FcValueListPtrU(*prev) == position)
 		    break;
 	    }
 	}
@@ -1113,7 +1269,7 @@ FcConfigAdd (FcValueList    **head,
 
 	if (FcDebug () & FC_DBG_EDIT)
 	{
-	    if (!*prev)
+	    if (!FcValueListPtrU(*prev))
 		printf ("position not on list\n");
 	}
     }
@@ -1127,12 +1283,12 @@ FcConfigAdd (FcValueList    **head,
     
     if (new)
     {
-	last = new;
-	while (last->next)
-	    last = last->next;
+	last = FcValueListPtrCreateDynamic(new);
+	while (FcValueListPtrU(FcValueListPtrU(last)->next))
+	    last = FcValueListPtrU(last)->next;
     
-	last->next = *prev;
-	*prev = new;
+	FcValueListPtrU(last)->next = *prev;
+	*prev = FcValueListPtrCreateDynamic(new);
     }
     
     if (FcDebug () & FC_DBG_EDIT)
@@ -1146,18 +1302,19 @@ FcConfigAdd (FcValueList    **head,
 }
 
 static void
-FcConfigDel (FcValueList    **head,
+FcConfigDel (FcValueListPtr *head,
 	     FcValueList    *position)
 {
-    FcValueList    **prev;
+    FcValueListPtr *prev;
 
-    for (prev = head; *prev; prev = &(*prev)->next)
+    for (prev = head; FcValueListPtrU(*prev); 
+	 prev = &(FcValueListPtrU(*prev)->next))
     {
-	if (*prev == position)
+	if (FcValueListPtrU(*prev) == position)
 	{
 	    *prev = position->next;
-	    position->next = 0;
-	    FcValueListDestroy (position);
+	    position->next = FcValueListPtrCreateDynamic(0);
+	    FcValueListDestroy (FcValueListPtrCreateDynamic(position));
 	    break;
 	}
     }
@@ -1189,8 +1346,8 @@ FcConfigPatternDel (FcPattern	*p,
     FcPatternElt    *e = FcPatternFindElt (p, object);
     if (!e)
 	return;
-    while (e->values)
-	FcConfigDel (&e->values, e->values);
+    while (FcValueListPtrU(e->values))
+	FcConfigDel (&e->values, FcValueListPtrU(e->values));
 }
 
 static void
@@ -1200,7 +1357,7 @@ FcConfigPatternCanon (FcPattern	    *p,
     FcPatternElt    *e = FcPatternFindElt (p, object);
     if (!e)
 	return;
-    if (!e->values)
+    if (!FcValueListPtrU(e->values))
 	FcPatternDel (p, object);
 }
 
@@ -1279,12 +1436,12 @@ FcConfigSubstituteWithPat (FcConfig    *config,
 	     * Check to see if there is a match, mark the location
 	     * to apply match-relative edits
 	     */
-	    st[i].value = FcConfigMatchValueList (m, t, st[i].elt->values);
+	    st[i].value = FcConfigMatchValueList (m, t, FcValueListPtrU(st[i].elt->values));
 	    if (!st[i].value)
 		break;
-	    if (t->qual == FcQualFirst && st[i].value != st[i].elt->values)
+	    if (t->qual == FcQualFirst && st[i].value != FcValueListPtrU(st[i].elt->values))
 		break;
-	    if (t->qual == FcQualNotFirst && st[i].value == st[i].elt->values)
+	    if (t->qual == FcQualNotFirst && st[i].value == FcValueListPtrU(st[i].elt->values))
 		break;
 	}
 	if (t)
@@ -1337,7 +1494,7 @@ FcConfigSubstituteWithPat (FcConfig    *config,
 		if (t)
 		{
 		    FcValueList	*thisValue = st[i].value;
-		    FcValueList	*nextValue = thisValue ? thisValue->next : 0;
+		    FcValueList	*nextValue = thisValue ? FcValueListPtrU(thisValue->next) : 0;
 		    
 		    /*
 		     * Append the new list of values after the current value
@@ -1503,7 +1660,9 @@ FcConfigFileExists (const FcChar8 *dir, const FcChar8 *file)
 #ifdef _WIN32
     if ((!path[0] || (path[strlen((char *) path)-1] != '/' &&
 		      path[strlen((char *) path)-1] != '\\')) &&
-	 (file[0] != '/' && file[0] != '\\'))
+	!(file[0] == '/' ||
+	  file[0] == '\\' ||
+	  (isalpha (file[0]) && file[1] == ':' && (file[2] == '/' || file[2] == '\\'))))
 	strcat ((char *) path, "\\");
 #else
     if ((!path[0] || path[strlen((char *) path)-1] != '/') && file[0] != '/')
@@ -1603,7 +1762,7 @@ FcConfigHome (void)
 	    home = getenv ("USERPROFILE");
 #endif
 
-	return home;
+	return (FcChar8 *) home;
     }
     return 0;
 }
@@ -1716,6 +1875,7 @@ FcConfigAppFontAddFile (FcConfig    *config,
 	}
 	FcStrListDone (sublist);
     }
+    FcStrSetDestroy (subdirs);
     return FcTrue;
 }
 
@@ -1763,6 +1923,7 @@ FcConfigAppFontAddDir (FcConfig	    *config,
 	}
 	FcStrListDone (sublist);
     }
+    FcStrSetDestroy (subdirs);
     return FcTrue;
 }
 
