@@ -1,7 +1,7 @@
 /*
  * $RCSId: xc/lib/fontconfig/src/fcdbg.c,v 1.10 2002/08/22 18:53:22 keithp Exp $
  *
- * Copyright  2000 Keith Packard
+ * Copyright © 2000 Keith Packard
  *
  * Permission to use, copy, modify, distribute, and sell this software and its
  * documentation for any purpose is hereby granted without fee, provided that
@@ -62,12 +62,12 @@ FcValuePrint (const FcValue v)
 }
 
 void
-FcValueListPrint (const FcValueList *l)
+FcValueListPrint (FcValueListPtr l)
 {
-    for (; l; l = l->next)
+    for (; FcValueListPtrU(l); l = FcValueListPtrU(l)->next)
     {
-	FcValuePrint (l->value);
-	switch (l->binding) {
+	FcValuePrint (FcValueCanonicalize(&FcValueListPtrU(l)->value));
+	switch (FcValueListPtrU(l)->binding) {
 	case FcValueBindingWeak:
 	    printf ("(w)");
 	    break;
@@ -89,9 +89,9 @@ FcLangSetPrint (const FcLangSet *ls)
     
     FcStrBufInit (&buf, init_buf, sizeof (init_buf));
     if (FcNameUnparseLangSet (&buf, ls) && FcStrBufChar (&buf,'\0'))
-	printf ("%s", buf.buf);
+       printf ("%s", buf.buf);
     else
-	printf ("langset (alloc error)");
+       printf ("langset (alloc error)");
     FcStrBufDestroy (&buf);
 }
 
@@ -106,12 +106,31 @@ FcPatternPrint (const FcPattern *p)
 	printf ("Null pattern\n");
 	return;
     }
-    printf ("Pattern %d of %d\n", p->num, p->size);
+    printf ("Pattern has %d elts (size %d)\n", p->num, p->size);
     for (i = 0; i < p->num; i++)
     {
-	e = &p->elts[i];
-	printf ("\t%s:", e->object);
-	FcValueListPrint (e->values);
+	e = FcPatternEltU(p->elts) + i;
+	printf ("\t%s:", FcObjectPtrU(e->object));
+	/* so that fc-match properly displays file: foo... */
+	if (e->object == FcObjectToPtr(FC_FILE))
+	{
+	    FcChar8 * s;
+	    FcPatternGetString (p, FC_FILE, 0, &s);
+	    printf (" \"%s\"", s);
+	    switch (FcValueListPtrU(e->values)->binding) {
+	    case FcValueBindingWeak:
+	        printf ("(w)");
+	        break;
+	    case FcValueBindingStrong:
+	        printf ("(s)");
+	        break;
+	    case FcValueBindingSame:
+	        printf ("(=)");
+	        break;
+	    }
+	}
+	else
+	    FcValueListPrint (e->values);
 	printf ("\n");
     }
     printf ("\n");
@@ -339,25 +358,19 @@ FcFontSetPrint (const FcFontSet *s)
     }
 }
 
-int
-FcDebug (void)
+int FcDebugVal;
+
+void
+FcInitDebug (void)
 {
-    static int  initialized;
-    static int  debug;
+    char    *e;
 
-    if (!initialized)
+    e = getenv ("FC_DEBUG");
+    if (e)
     {
-	char    *e;
-
-	initialized = 1;
-	e = getenv ("FC_DEBUG");
-	if (e)
-	{
-	    printf ("FC_DEBUG=%s\n", e);
-	    debug = atoi (e);
-	    if (debug < 0)
-		debug = 0;
-	}
+        printf ("FC_DEBUG=%s\n", e);
+        FcDebugVal = atoi (e);
+        if (FcDebugVal < 0)
+   	    FcDebugVal = 0;
     }
-    return debug;
 }
