@@ -1,7 +1,7 @@
 /*
  * $RCSId: xc/lib/fontconfig/src/fcxml.c,v 1.21 2002/08/22 18:53:22 keithp Exp $
  *
- * Copyright  2002 Keith Packard
+ * Copyright © 2002 Keith Packard
  *
  * Permission to use, copy, modify, distribute, and sell this software and its
  * documentation for any purpose is hereby granted without fee, provided that
@@ -22,8 +22,23 @@
  * PERFORMANCE OF THIS SOFTWARE.
  */
 
+#include <fcntl.h>
 #include <stdarg.h>
 #include "fcint.h"
+#include <dirent.h>
+
+#if ENABLE_LIBXML2
+
+#include <libxml/parser.h>
+
+#define XML_Char			xmlChar
+#define XML_Parser			xmlParserCtxtPtr
+#define XML_ParserFree			xmlFreeParserCtxt
+#define XML_GetCurrentLineNumber	xmlSAX2GetLineNumber
+#define XML_GetErrorCode		xmlCtxtGetLastError
+#define XML_ErrorString(Error)		(Error)->message
+
+#else /* ENABLE_LIBXML2 */
 
 #ifndef HAVE_XMLPARSE_H
 #define HAVE_XMLPARSE_H 0
@@ -35,33 +50,14 @@
 #include <expat.h>
 #endif
 
+#endif /* ENABLE_LIBXML2 */
+
 #ifdef _WIN32
 #define STRICT
 #include <windows.h>
 #undef STRICT
 #endif
 
-FcTest *
-FcTestCreate (FcMatchKind   kind, 
-	      FcQual	    qual,
-	      const FcChar8 *field,
-	      FcOp	    compare,
-	      FcExpr	    *expr)
-{
-    FcTest	*test = (FcTest *) malloc (sizeof (FcTest));
-
-    if (test)
-    {
-	FcMemAlloc (FC_MEM_TEST, sizeof (FcTest));
-	test->next = 0;
-	test->kind = kind;
-	test->qual = qual;
-	test->field = (char *) FcStrCopy (field);
-	test->op = compare;
-	test->expr = expr;
-    }
-    return test;
-}
 
 void
 FcTestDestroy (FcTest *test)
@@ -268,22 +264,6 @@ FcExprDestroy (FcExpr *e)
     free (e);
 }
 
-FcEdit *
-FcEditCreate (const char *field, FcOp op, FcExpr *expr, FcValueBinding binding)
-{
-    FcEdit *e = (FcEdit *) malloc (sizeof (FcEdit));
-
-    if (e)
-    {
-	e->next = 0;
-	e->field = field;   /* already saved in grammar */
-	e->op = op;
-	e->expr = expr;
-	e->binding = binding;
-    }
-    return e;
-}
-
 void
 FcEditDestroy (FcEdit *e)
 {
@@ -323,6 +303,8 @@ typedef enum _FcElement {
     FcElementAcceptfont,
     FcElementRejectfont,
     FcElementGlob,
+    FcElementPattern,
+    FcElementPatelt,
 
     FcElementTest,
     FcElementEdit,
@@ -357,70 +339,72 @@ typedef enum _FcElement {
     FcElementUnknown
 } FcElement;
 
+static struct {
+    const char  name[16];
+    FcElement   element;
+} fcElementMap[] = {
+    { "fontconfig",	FcElementFontconfig },
+    { "dir",		FcElementDir },
+    { "cache",		FcElementCache },
+    { "include",	FcElementInclude },
+    { "config",		FcElementConfig },
+    { "match",		FcElementMatch },
+    { "alias",		FcElementAlias },
+    
+    { "blank",		FcElementBlank },
+    { "rescan",		FcElementRescan },
+
+    { "prefer",		FcElementPrefer },
+    { "accept",		FcElementAccept },
+    { "default",	FcElementDefault },
+    { "family",		FcElementFamily },
+
+    { "selectfont",	FcElementSelectfont },
+    { "acceptfont",	FcElementAcceptfont },
+    { "rejectfont",	FcElementRejectfont },
+    { "glob",		FcElementGlob },
+    { "pattern",	FcElementPattern },
+    { "patelt",		FcElementPatelt },
+
+    { "test",		FcElementTest },
+    { "edit",		FcElementEdit },
+    { "int",		FcElementInt },
+    { "double",		FcElementDouble },
+    { "string",		FcElementString },
+    { "matrix",		FcElementMatrix },
+    { "bool",		FcElementBool },
+    { "charset",	FcElementCharset },
+    { "name",		FcElementName },
+    { "const",		FcElementConst },
+    { "or",		FcElementOr },
+    { "and",		FcElementAnd },
+    { "eq",		FcElementEq },
+    { "not_eq",		FcElementNotEq },
+    { "less",		FcElementLess },
+    { "less_eq",	FcElementLessEq },
+    { "more",		FcElementMore },
+    { "more_eq",	FcElementMoreEq },
+    { "contains",	FcElementContains },
+    { "not_contains",	FcElementNotContains },
+    { "plus",		FcElementPlus },
+    { "minus",		FcElementMinus },
+    { "times",		FcElementTimes },
+    { "divide",		FcElementDivide },
+    { "not",		FcElementNot },
+    { "if",		FcElementIf },
+    { "floor",		FcElementFloor },
+    { "ceil",		FcElementCeil },
+    { "round",		FcElementRound },
+    { "trunc",		FcElementTrunc },
+};
+#define NUM_ELEMENT_MAPS (int) (sizeof fcElementMap / sizeof fcElementMap[0])
+
 static FcElement
 FcElementMap (const XML_Char *name)
 {
-    static struct {
-	char	    *name;
-	FcElement   element;
-    } fcElementMap[] = {
-	{ "fontconfig",	FcElementFontconfig },
-	{ "dir",	FcElementDir },
-	{ "cache",	FcElementCache },
-	{ "include",	FcElementInclude },
-	{ "config",	FcElementConfig },
-	{ "match",	FcElementMatch },
-	{ "alias",	FcElementAlias },
-	
-	{ "blank",	FcElementBlank },
-	{ "rescan",	FcElementRescan },
-
-	{ "prefer",	FcElementPrefer },
-	{ "accept",	FcElementAccept },
-	{ "default",	FcElementDefault },
-	{ "family",	FcElementFamily },
-
-	{ "selectfont",	FcElementSelectfont },
-	{ "acceptfont",	FcElementAcceptfont },
-	{ "rejectfont",	FcElementRejectfont },
-	{ "glob",	FcElementGlob },
-
-	{ "test",	FcElementTest },
-	{ "edit",	FcElementEdit },
-	{ "int",	FcElementInt },
-	{ "double",	FcElementDouble },
-	{ "string",	FcElementString },
-	{ "matrix",	FcElementMatrix },
-	{ "bool",	FcElementBool },
-	{ "charset",	FcElementCharset },
-	{ "name",	FcElementName },
-	{ "const",	FcElementConst },
-	{ "or",		FcElementOr },
-	{ "and",	FcElementAnd },
-	{ "eq",		FcElementEq },
-	{ "not_eq",	FcElementNotEq },
-	{ "less",	FcElementLess },
-	{ "less_eq",	FcElementLessEq },
-	{ "more",	FcElementMore },
-	{ "more_eq",	FcElementMoreEq },
-	{ "contains",	FcElementContains },
-	{ "not_contains",FcElementNotContains },
-	{ "plus",	FcElementPlus },
-	{ "minus",	FcElementMinus },
-	{ "times",	FcElementTimes },
-	{ "divide",	FcElementDivide },
-	{ "not",	FcElementNot },
-	{ "if",		FcElementIf },
-	{ "floor",	FcElementFloor },
-	{ "ceil",	FcElementCeil },
-	{ "round",	FcElementRound },
-	{ "trunc",	FcElementTrunc },
-	
-	{ 0,		0 }
-    };
 
     int	    i;
-    for (i = 0; fcElementMap[i].name; i++)
+    for (i = 0; i < NUM_ELEMENT_MAPS; i++)
 	if (!strcmp ((char *) name, fcElementMap[i].name))
 	    return fcElementMap[i].element;
     return FcElementUnknown;
@@ -441,6 +425,7 @@ typedef enum _FcVStackTag {
     FcVStackField,
     FcVStackConstant,
     FcVStackGlob,
+    FcVStackPattern,
     
     FcVStackPrefer,
     FcVStackAccept,
@@ -473,6 +458,8 @@ typedef struct _FcVStack {
 	FcOp		op;
 	FcExpr		*expr;
 	FcEdit		*edit;
+
+	FcPattern	*pattern;
     } u;
 } FcVStack;
 
@@ -490,9 +477,9 @@ typedef enum _FcConfigSeverity {
 } FcConfigSeverity;
 
 static void
-FcConfigMessage (FcConfigParse *parse, FcConfigSeverity severe, char *fmt, ...)
+FcConfigMessage (FcConfigParse *parse, FcConfigSeverity severe, const char *fmt, ...)
 {
-    char	*s = "unknown";
+    const char	*s = "unknown";
     va_list	args;
 
     va_start (args, fmt);
@@ -506,10 +493,10 @@ FcConfigMessage (FcConfigParse *parse, FcConfigSeverity severe, char *fmt, ...)
     {
 	if (parse->name)
 	    fprintf (stderr, "Fontconfig %s: \"%s\", line %d: ", s,
-		     parse->name, XML_GetCurrentLineNumber (parse->parser));
+		     parse->name, (int)XML_GetCurrentLineNumber (parse->parser));
 	else
 	    fprintf (stderr, "Fontconfig %s: line %d: ", s,
-		     XML_GetCurrentLineNumber (parse->parser));
+		     (int)XML_GetCurrentLineNumber (parse->parser));
 	if (severe >= FcSevereError)
 	    parse->error = FcTrue;
     }
@@ -520,6 +507,191 @@ FcConfigMessage (FcConfigParse *parse, FcConfigSeverity severe, char *fmt, ...)
     va_end (args);
 }
 
+
+static const char *
+FcTypeName (FcType type)
+{
+    switch (type) {
+    case FcTypeVoid:
+	return "void";
+    case FcTypeInteger:
+    case FcTypeDouble:
+	return "number";
+    case FcTypeString:
+	return "string";
+    case FcTypeBool:
+	return "bool";
+    case FcTypeMatrix:
+	return "matrix";
+    case FcTypeCharSet:
+	return "charset";
+    case FcTypeFTFace:
+	return "FT_Face";
+    case FcTypeLangSet:
+	return "langset";
+    default:
+	return "unknown";
+    }
+}
+
+static void
+FcTypecheckValue (FcConfigParse *parse, FcType value, FcType type)
+{
+    if (value == FcTypeInteger)
+	value = FcTypeDouble;
+    if (type == FcTypeInteger)
+	type = FcTypeDouble;
+    if (value != type)
+    {
+	if ((value == FcTypeLangSet && type == FcTypeString) ||
+	    (value == FcTypeString && type == FcTypeLangSet))
+	    return;
+	FcConfigMessage (parse, FcSevereWarning, "saw %s, expected %s",
+			 FcTypeName (value), FcTypeName (type));
+    }
+}
+
+static void
+FcTypecheckExpr (FcConfigParse *parse, FcExpr *expr, FcType type)
+{
+    const FcObjectType	*o;
+    const FcConstant	*c;
+    
+    switch (expr->op) {
+    case FcOpInteger:
+    case FcOpDouble:
+	FcTypecheckValue (parse, FcTypeDouble, type);
+	break;
+    case FcOpString:
+	FcTypecheckValue (parse, FcTypeString, type);
+	break;
+    case FcOpMatrix:
+	FcTypecheckValue (parse, FcTypeMatrix, type);
+	break;
+    case FcOpBool:
+	FcTypecheckValue (parse, FcTypeBool, type);
+	break;
+    case FcOpCharSet:
+	FcTypecheckValue (parse, FcTypeCharSet, type);
+	break;
+    case FcOpNil:
+	break;
+    case FcOpField:
+	o = FcNameGetObjectType (expr->u.field);
+	if (o)
+	    FcTypecheckValue (parse, o->type, type);
+	break;
+    case FcOpConst:
+	c = FcNameGetConstant (expr->u.constant);
+	if (c)
+	{
+	    o = FcNameGetObjectType (c->object);
+	    if (o)
+		FcTypecheckValue (parse, o->type, type);
+	}
+        else 
+            FcConfigMessage (parse, FcSevereWarning, 
+                             "invalid constant used : %s",
+                             expr->u.constant);
+	break;
+    case FcOpQuest:
+	FcTypecheckExpr (parse, expr->u.tree.left, FcTypeBool);
+	FcTypecheckExpr (parse, expr->u.tree.right->u.tree.left, type);
+	FcTypecheckExpr (parse, expr->u.tree.right->u.tree.right, type);
+	break;
+    case FcOpAssign:
+    case FcOpAssignReplace:
+	break;
+    case FcOpEqual:
+    case FcOpNotEqual:
+    case FcOpLess:
+    case FcOpLessEqual:
+    case FcOpMore:
+    case FcOpMoreEqual:
+    case FcOpContains:
+    case FcOpNotContains:
+    case FcOpListing:
+	FcTypecheckValue (parse, FcTypeBool, type);
+	break;
+    case FcOpComma:
+    case FcOpOr:
+    case FcOpAnd:
+    case FcOpPlus:
+    case FcOpMinus:
+    case FcOpTimes:
+    case FcOpDivide:
+	FcTypecheckExpr (parse, expr->u.tree.left, type);
+	FcTypecheckExpr (parse, expr->u.tree.right, type);
+	break;
+    case FcOpNot:
+	FcTypecheckValue (parse, FcTypeBool, type);
+	FcTypecheckExpr (parse, expr->u.tree.left, FcTypeBool);
+	break;
+    case FcOpFloor:
+    case FcOpCeil:
+    case FcOpRound:
+    case FcOpTrunc:
+	FcTypecheckValue (parse, FcTypeDouble, type);
+	FcTypecheckExpr (parse, expr->u.tree.left, FcTypeDouble);
+	break;
+    default:
+	break;
+    }
+}
+
+static FcTest *
+FcTestCreate (FcConfigParse *parse,
+	      FcMatchKind   kind, 
+	      FcQual	    qual,
+	      const FcChar8 *field,
+	      FcOp	    compare,
+	      FcExpr	    *expr)
+{
+    FcTest	*test = (FcTest *) malloc (sizeof (FcTest));
+
+    if (test)
+    {
+	const FcObjectType	*o;
+	
+	FcMemAlloc (FC_MEM_TEST, sizeof (FcTest));
+	test->next = 0;
+	test->kind = kind;
+	test->qual = qual;
+	test->field = (char *) FcStrCopy (field);
+	test->op = compare;
+	test->expr = expr;
+	o = FcNameGetObjectType (test->field);
+	if (o)
+	    FcTypecheckExpr (parse, expr, o->type);
+    }
+    return test;
+}
+
+static FcEdit *
+FcEditCreate (FcConfigParse	*parse,
+	      const char	*field,
+	      FcOp		op,
+	      FcExpr		*expr,
+	      FcValueBinding	binding)
+{
+    FcEdit *e = (FcEdit *) malloc (sizeof (FcEdit));
+
+    if (e)
+    {
+	const FcObjectType	*o;
+
+	e->next = 0;
+	e->field = field;   /* already saved in grammar */
+	e->op = op;
+	e->expr = expr;
+	e->binding = binding;
+	o = FcNameGetObjectType (e->field);
+	if (o)
+	    FcTypecheckExpr (parse, expr, o->type);
+    }
+    return e;
+}
+
 static void
 FcVStackPush (FcConfigParse *parse, FcVStack *vstack)
 {
@@ -560,6 +732,9 @@ FcVStackDestroy (FcVStack *vstack)
 	case FcVStackGlob:
 	    FcStrFree (vstack->u.string);
 	    break;
+	case FcVStackPattern:
+	    FcPatternDestroy (vstack->u.pattern);
+	    break;
 	case FcVStackInteger:
 	case FcVStackDouble:
 	    break;
@@ -688,6 +863,18 @@ FcVStackPushEdit (FcConfigParse *parse, FcEdit *edit)
     return FcTrue;
 }
 
+static FcBool
+FcVStackPushPattern (FcConfigParse *parse, FcPattern *pattern)
+{
+    FcVStack    *vstack = FcVStackCreate ();
+    if (!vstack)
+	return FcFalse;
+    vstack->u.pattern = pattern;
+    vstack->tag = FcVStackPattern;
+    FcVStackPush (parse, vstack);
+    return FcTrue;
+}
+
 static FcVStack *
 FcVStackFetch (FcConfigParse *parse, int off)
 {
@@ -737,7 +924,6 @@ FcVStackElements (FcConfigParse *parse)
 static FcChar8 **
 FcConfigSaveAttr (const XML_Char **attr)
 {
-    int		n;
     int		slen;
     int		i;
     FcChar8	**new;
@@ -747,8 +933,7 @@ FcConfigSaveAttr (const XML_Char **attr)
 	return 0;
     slen = 0;
     for (i = 0; attr[i]; i++)
-	slen += strlen (attr[i]) + 1;
-    n = i;
+	slen += strlen ((char *) attr[i]) + 1;
     new = malloc ((i + 1) * sizeof (FcChar8 *) + slen);
     if (!new)
 	return 0;
@@ -831,13 +1016,16 @@ FcConfigCleanup (FcConfigParse	*parse)
 }
 
 static const FcChar8 *
-FcConfigGetAttribute (FcConfigParse *parse, char *attr)
+FcConfigGetAttribute (FcConfigParse *parse, const char *attr)
 {
     FcChar8 **attrs;
     if (!parse->pstack)
 	return 0;
 
     attrs = parse->pstack->attr;
+    if (!attrs)
+        return 0;
+
     while (*attrs)
     {
 	if (!strcmp ((char *) *attrs, attr))
@@ -958,7 +1146,7 @@ FcStrtod (char *s, char **end)
 	int	slen = strlen (s);
 	int	dlen = strlen (locale_data->decimal_point);
 	
-	if (slen + dlen > sizeof (buf))
+	if (slen + dlen > (int) sizeof (buf))
 	{
 	    if (end)
 		*end = s;
@@ -1068,15 +1256,14 @@ FcParseMatrix (FcConfigParse *parse)
 }
 
 static FcBool
-FcConfigLexBool (const FcChar8 *bool)
+FcConfigLexBool (FcConfigParse *parse, const FcChar8 *bool)
 {
-    if (*bool == 't' || *bool == 'T')
-	return FcTrue;
-    if (*bool == 'y' || *bool == 'Y')
-	return FcTrue;
-    if (*bool == '1')
-	return FcTrue;
-    return FcFalse;
+    FcBool  result = FcFalse;
+
+    if (!FcNameBool (bool, &result))
+	FcConfigMessage (parse, FcSevereWarning, "\"%s\" is not known boolean",
+			 bool);
+    return result;
 }
 
 static void
@@ -1092,7 +1279,7 @@ FcParseBool (FcConfigParse *parse)
 	FcConfigMessage (parse, FcSevereError, "out of memory");
 	return;
     }
-    FcVStackPushBool (parse, FcConfigLexBool (s));
+    FcVStackPushBool (parse, FcConfigLexBool (parse, s));
     FcStrFree (s);
 }
 
@@ -1133,8 +1320,7 @@ FcParseFamilies (FcConfigParse *parse, FcVStackTag tag)
 	if (!FcVStackPushExpr (parse, tag, expr))
 	{
 	    FcConfigMessage (parse, FcSevereError, "out of memory");
-	    if (expr)
-		FcExprDestroy (expr);
+            FcExprDestroy (expr);
 	}
     }
 }
@@ -1224,7 +1410,8 @@ FcParseAlias (FcConfigParse *parse)
     }
     if (prefer)
     {
-	edit = FcEditCreate (FcConfigSaveField ("family"),
+	edit = FcEditCreate (parse, 
+			     FcConfigSaveField ("family"),
 			     FcOpPrepend,
 			     prefer,
 			     FcValueBindingWeak);
@@ -1236,7 +1423,8 @@ FcParseAlias (FcConfigParse *parse)
     if (accept)
     {
 	next = edit;
-	edit = FcEditCreate (FcConfigSaveField ("family"),
+	edit = FcEditCreate (parse,
+			     FcConfigSaveField ("family"),
 			     FcOpAppend,
 			     accept,
 			     FcValueBindingWeak);
@@ -1248,7 +1436,8 @@ FcParseAlias (FcConfigParse *parse)
     if (def)
     {
 	next = edit;
-	edit = FcEditCreate (FcConfigSaveField ("family"),
+	edit = FcEditCreate (parse,
+			     FcConfigSaveField ("family"),
 			     FcOpAppendLast,
 			     def,
 			     FcValueBindingWeak);
@@ -1259,7 +1448,7 @@ FcParseAlias (FcConfigParse *parse)
     }
     if (edit)
     {
-	test = FcTestCreate (FcMatchPattern,
+	test = FcTestCreate (parse, FcMatchPattern,
 			     FcQualAny,
 			     (FcChar8 *) FC_FAMILY,
 			     FcOpEqual,
@@ -1321,6 +1510,8 @@ FcPopExpr (FcConfigParse *parse)
 	break;
     case FcVStackEdit:
 	break;
+    default:
+	break;
     }
     FcVStackDestroy (vstack);
     return expr;
@@ -1412,7 +1603,7 @@ FcParseInclude (FcConfigParse *parse)
 	return;
     }
     i = FcConfigGetAttribute (parse, "ignore_missing");
-    if (i && FcConfigLexBool ((FcChar8 *) i) == FcTrue)
+    if (i && FcConfigLexBool (parse, (FcChar8 *) i) == FcTrue)
 	ignore_missing = FcTrue;
     if (!FcConfigParseAndLoad (parse->config, s, !ignore_missing))
 	parse->error = FcTrue;
@@ -1420,7 +1611,7 @@ FcParseInclude (FcConfigParse *parse)
 }
 
 typedef struct _FcOpMap {
-    char    *name;
+    char    name[16];
     FcOp    op;
 } FcOpMap;
 
@@ -1446,7 +1637,7 @@ static const FcOpMap fcCompareOps[] = {
     { "not_contains",	FcOpNotContains	    }
 };
 
-#define NUM_COMPARE_OPS (sizeof fcCompareOps / sizeof fcCompareOps[0])
+#define NUM_COMPARE_OPS	(int) (sizeof fcCompareOps / sizeof fcCompareOps[0])
 
 static FcOp
 FcConfigLexCompare (const FcChar8 *compare)
@@ -1528,7 +1719,7 @@ FcParseTest (FcConfigParse *parse)
 	FcConfigMessage (parse, FcSevereWarning, "missing test expression");
 	return;
     }
-    test = FcTestCreate (kind, qual, name, compare, expr);
+    test = FcTestCreate (parse, kind, qual, name, compare, expr);
     if (!test)
     {
 	FcConfigMessage (parse, FcSevereError, "out of memory");
@@ -1546,7 +1737,7 @@ static const FcOpMap fcModeOps[] = {
     { "append_last",	FcOpAppendLast	    },
 };
 
-#define NUM_MODE_OPS (sizeof fcModeOps / sizeof fcModeOps[0])
+#define NUM_MODE_OPS (int) (sizeof fcModeOps / sizeof fcModeOps[0])
 
 static FcOp
 FcConfigLexMode (const FcChar8 *mode)
@@ -1601,7 +1792,7 @@ FcParseEdit (FcConfigParse *parse)
 	}
     }
     expr = FcPopBinary (parse, FcOpComma);
-    edit = FcEditCreate ((char *) FcStrCopy (name), mode, expr, binding);
+    edit = FcEditCreate (parse, (char *) FcStrCopy (name), mode, expr, binding);
     if (!edit)
     {
 	FcConfigMessage (parse, FcSevereError, "out of memory");
@@ -1675,6 +1866,16 @@ FcParseAcceptRejectFont (FcConfigParse *parse, FcElement element)
 		FcConfigMessage (parse, FcSevereError, "out of memory");
 	    }
 	    break;
+	case FcVStackPattern:
+	    if (!FcConfigPatternsAdd (parse->config,
+				      vstack->u.pattern,
+				      element == FcElementAcceptfont))
+	    {
+		FcConfigMessage (parse, FcSevereError, "out of memory");
+	    }
+	    else
+		vstack->tag = FcVStackNone;
+	    break;
 	default:
 	    FcConfigMessage (parse, FcSevereWarning, "bad font selector");
 	    break;
@@ -1683,6 +1884,123 @@ FcParseAcceptRejectFont (FcConfigParse *parse, FcElement element)
     }
 }
 
+
+static FcValue
+FcPopValue (FcConfigParse *parse)
+{
+    FcVStack	*vstack = FcVStackPop (parse);
+    FcValue	value;
+    
+    value.type = FcTypeVoid;
+    
+    if (!vstack)
+	return value;
+    
+    switch (vstack->tag) {
+    case FcVStackString:
+	value.u.s = FcStrCopy (vstack->u.string);
+	if (value.u.s)
+	    value.type = FcTypeString;
+	break;
+    case FcVStackConstant:
+	if (FcNameConstant (vstack->u.string, &value.u.i))
+	    value.type = FcTypeInteger;
+	break;
+    case FcVStackInteger:
+	value.u.i = vstack->u.integer;
+	value.type = FcTypeInteger;
+	break;
+    case FcVStackDouble:
+	value.u.d = vstack->u._double;
+	value.type = FcTypeInteger;
+	break;
+    case FcVStackMatrix:
+	value.u.m = FcMatrixCopy (vstack->u.matrix);
+	if (value.u.m)
+	    value.type = FcTypeMatrix;
+	break;
+    case FcVStackBool:
+	value.u.b = vstack->u.bool;
+	value.type = FcTypeBool;
+	break;
+    default:
+	FcConfigMessage (parse, FcSevereWarning, "unknown pattern element %d", 
+			 vstack->tag);
+	break;
+    }
+    FcVStackDestroy (vstack);
+    
+    return value;
+}
+
+static void
+FcParsePatelt (FcConfigParse *parse)
+{
+    FcValue	value;
+    FcPattern	*pattern = FcPatternCreate ();
+    const char	*name;
+
+    if (!pattern)
+    {
+	FcConfigMessage (parse, FcSevereError, "out of memory");
+	return;
+    }
+
+    name = (char *) FcConfigGetAttribute (parse, "name");
+    if (!name)
+    {
+	FcConfigMessage (parse, FcSevereWarning, "missing pattern element name");
+	return;
+    }
+    
+    for (;;)
+    {
+	value = FcPopValue (parse);
+	if (value.type == FcTypeVoid)
+	    break;
+	if (!FcPatternAdd (pattern, name, value, FcTrue))
+	{
+	    FcConfigMessage (parse, FcSevereError, "out of memory");
+	    break;
+	}
+    }
+
+    FcVStackPushPattern (parse, FcPatternFreeze(pattern));
+    FcPatternDestroy (pattern);
+}
+
+static void
+FcParsePattern (FcConfigParse *parse)
+{
+    FcVStack	*vstack;
+    FcPattern	*pattern = FcPatternCreate ();
+
+    if (!pattern)
+    {
+	FcConfigMessage (parse, FcSevereError, "out of memory");
+	return;
+    }
+	
+    while ((vstack = FcVStackPop (parse)))
+    {
+	switch (vstack->tag) {
+	case FcVStackPattern:
+	    if (!FcPatternAppend (pattern, vstack->u.pattern))
+	    {
+		FcConfigMessage (parse, FcSevereError, "out of memory");
+		return;
+	    }
+	    break;
+	default:
+	    FcConfigMessage (parse, FcSevereWarning, "unknown pattern element");
+	    break;
+	}
+	FcVStackDestroy (vstack);
+    }
+
+    FcVStackPushPattern (parse, pattern);
+}
+
 static void
 FcEndElement(void *userData, const XML_Char *name)
 {
@@ -1730,7 +2048,7 @@ FcEndElement(void *userData, const XML_Char *name)
 	if (!FcStrUsesHome (data) || FcConfigHome ())
 	{
 	    if (!FcConfigAddDir (parse->config, data))
-		FcConfigMessage (parse, FcSevereError, "out of memory");
+		FcConfigMessage (parse, FcSevereError, "out of memory; cannot add directory %s", data);
 	}
 	FcStrFree (data);
 	break;
@@ -1814,6 +2132,12 @@ FcEndElement(void *userData, const XML_Char *name)
     case FcElementGlob:
 	FcParseString (parse, FcVStackGlob);
 	break;
+    case FcElementPattern:
+	FcParsePattern (parse);
+	break;
+    case FcElementPatelt:
+	FcParsePatelt (parse);
+	break;
     case FcElementName:
 	FcParseString (parse, FcVStackField);
 	break;
@@ -1910,11 +2234,118 @@ FcStartDoctypeDecl (void	    *userData,
 	FcConfigMessage (parse, FcSevereError, "invalid doctype \"%s\"", doctypeName);
 }
 
+#if ENABLE_LIBXML2
+
+static void
+FcInternalSubsetDecl (void            *userData,
+		      const XML_Char  *doctypeName,
+		      const XML_Char  *sysid,
+		      const XML_Char  *pubid)
+{
+    FcStartDoctypeDecl (userData, doctypeName, sysid, pubid, 1);
+}
+
+static void
+FcExternalSubsetDecl (void            *userData,
+		      const XML_Char  *doctypeName,
+		      const XML_Char  *sysid,
+		      const XML_Char  *pubid)
+{
+    FcStartDoctypeDecl (userData, doctypeName, sysid, pubid, 0);
+}
+
+#else /* ENABLE_LIBXML2 */
+
 static void
 FcEndDoctypeDecl (void *userData)
 {
 }
 
+#endif /* ENABLE_LIBXML2 */
+
+static FcBool
+FcConfigParseAndLoadDir (FcConfig	*config,
+			 const FcChar8	*name,
+			 const FcChar8	*dir,
+			 FcBool		complain)
+{
+    DIR		    *d;
+    struct dirent   *e;
+    FcBool	    ret = FcTrue;
+    FcChar8	    *file;
+    FcChar8	    *base;
+    FcStrSet	    *files;
+
+    d = opendir ((char *) dir);
+    if (!d)
+    {
+	if (complain)
+	    FcConfigMessage (0, FcSevereError, "Cannot open config dir \"%s\"",
+			     name);
+	ret = FcFalse;
+	goto bail0;
+    }
+    /* freed below */
+    file = (FcChar8 *) malloc (strlen ((char *) dir) + 1 + FC_MAX_FILE_LEN + 1);
+    if (!file)
+    {
+	ret = FcFalse;
+	goto bail1;
+    }
+    
+    strcpy ((char *) file, (char *) dir);
+    strcat ((char *) file, "/");
+    base = file + strlen ((char *) file);
+    
+    files = FcStrSetCreate ();
+    if (!files)
+    {
+	ret = FcFalse;
+	goto bail2;
+    }
+    
+    if (FcDebug () & FC_DBG_CONFIG)
+	printf ("\tScanning config dir %s\n", dir);
+	
+    while (ret && (e = readdir (d)))
+    {
+	int d_len;
+#define TAIL	    ".conf"
+#define TAIL_LEN    5
+	/*
+	 * Add all files of the form [0-9]*.conf
+	 */
+	if ('0' <= e->d_name[0] && e->d_name[0] <= '9' &&
+	    (d_len = strlen (e->d_name)) < FC_MAX_FILE_LEN &&
+	    d_len > TAIL_LEN &&
+	    strcmp (e->d_name + d_len - TAIL_LEN, TAIL) == 0)
+	{
+	    strcpy ((char *) base, (char *) e->d_name);
+	    if (!FcStrSetAdd (files, file))
+	    {
+		ret = FcFalse;
+		goto bail3;
+	    }
+	}
+    }
+    if (ret)
+    {
+	int i;
+	qsort (files->strs, files->num, sizeof (FcChar8 *), 
+	       (int (*)(const void *, const void *)) FcStrCmp);
+	for (i = 0; ret && i < files->num; i++)
+	    ret = FcConfigParseAndLoad (config, files->strs[i], complain);
+    }
+bail3:
+    FcStrSetDestroy (files);
+bail2:
+    free (file);
+bail1:
+    closedir (d);
+bail0:
+    return ret || !complain;
+}
+
 FcBool
 FcConfigParseAndLoad (FcConfig	    *config,
 		      const FcChar8 *name,
@@ -1923,54 +2354,102 @@ FcConfigParseAndLoad (FcConfig	    *config,
 
     XML_Parser	    p;
     FcChar8	    *filename;
-    FILE	    *f;
+    int		    fd;
     int		    len;
-    void	    *buf;
     FcConfigParse   parse;
     FcBool	    error = FcTrue;
     
+#if ENABLE_LIBXML2
+    xmlSAXHandler   sax;
+    char            buf[BUFSIZ];
+#else
+    void	    *buf;
+#endif
+    
     filename = FcConfigFilename (name);
     if (!filename)
 	goto bail0;
     
+    if (FcStrSetMember (config->configFiles, filename))
+    {
+        FcStrFree (filename);
+        return FcTrue;
+    }
+
     if (!FcStrSetAdd (config->configFiles, filename))
     {
 	FcStrFree (filename);
 	goto bail0;
     }
 
-    f = fopen ((char *) filename, "r");
-    FcStrFree (filename);
-    if (!f)
+    if (FcFileIsDir (filename))
+    {
+	FcBool ret = FcConfigParseAndLoadDir (config, name, filename, complain);
+	FcStrFree (filename);
+	return ret;
+    }
+
+    if (FcDebug () & FC_DBG_CONFIG)
+	printf ("\tLoading config file %s\n", filename);
+
+    fd = open ((char *) filename, O_RDONLY);
+    if (fd == -1) { 
+	FcStrFree (filename);
 	goto bail0;
+    }
     
+#if ENABLE_LIBXML2
+    memset(&sax, 0, sizeof(sax));
+
+    sax.internalSubset = FcInternalSubsetDecl;
+    sax.externalSubset = FcExternalSubsetDecl;
+    sax.startElement = FcStartElement;
+    sax.endElement = FcEndElement;
+    sax.characters = FcCharacterData;
+
+    p = xmlCreatePushParserCtxt (&sax, &parse, NULL, 0, (const char *) filename);
+#else
     p = XML_ParserCreate ("UTF-8");
+#endif
+    FcStrFree (filename);
+
     if (!p)
 	goto bail1;
 
     if (!FcConfigInit (&parse, name, config, p))
 	goto bail2;
 
+#if !ENABLE_LIBXML2
+
     XML_SetUserData (p, &parse);
     
     XML_SetDoctypeDeclHandler (p, FcStartDoctypeDecl, FcEndDoctypeDecl);
     XML_SetElementHandler (p, FcStartElement, FcEndElement);
     XML_SetCharacterDataHandler (p, FcCharacterData);
 	
+#endif /* ENABLE_LIBXML2 */
+
     do {
+#if !ENABLE_LIBXML2
 	buf = XML_GetBuffer (p, BUFSIZ);
 	if (!buf)
 	{
 	    FcConfigMessage (&parse, FcSevereError, "cannot get parse buffer");
 	    goto bail3;
 	}
-	len = fread (buf, 1, BUFSIZ, f);
+#endif
+	len = read (fd, buf, BUFSIZ);
 	if (len < 0)
 	{
 	    FcConfigMessage (&parse, FcSevereError, "failed reading config file");
 	    goto bail3;
 	}
+
+#if ENABLE_LIBXML2
+	if (xmlParseChunk (p, buf, len, len == 0))
+#else
 	if (!XML_ParseBuffer (p, len, len == 0))
+#endif
 	{
 	    FcConfigMessage (&parse, FcSevereError, "%s", 
 			   XML_ErrorString (XML_GetErrorCode (p)));
@@ -1983,7 +2462,8 @@ bail3:
 bail2:
     XML_ParserFree (p);
 bail1:
-    fclose (f);
+    close (fd);
+    fd = -1;
 bail0:
     if (error && complain)
     {
